#include <Wire.h>
#include <LSM6.h>

const int LED_PIN=13;

const int BLOCK_SIZE = 7;
const int BLOCKS_TO_READ = 5;

const unsigned long KEEP_RUNNING_TIMEOUT = 5000UL;
const unsigned long NO_DATA_TIMEOUT = 250UL;
const unsigned long READ_CHUNK_PERIOD = 250UL;

const int COMMAND_STOP_READING_IMU = 0;
const int COMMAND_START_READING_IMU = 1;
const int COMMAND_KEEP_READING_IMU = 2;
const int COMMAND_GET_IDENTIFICATION_CODE = 3;

//Control Block Status
const int CBS_STATISTICS = 0;
const int CBS_START_DATA_STREAM = 1;
const int CBS_END_DATA_STREAM = 2;
const int CBS_ERROR_IMU_NOT_FOUND = 3;
const int CBS_ERROR_IMU_FIFO_OVERFLOW = 4;
const int CBS_ERROR_IMU_FIFO_READ_LESS_THAN_REQUESTED = 5;
const int CBS_ERROR_IMU_FIFO_PATTERN = 6;
const int CBS_ERROR_IMU_FIFO_NO_DATA_TIMEOUT = 7;
const int CBS_ERROR_WRONG_COMMAND = 8;
const int CBS_ERROR_KEEP_RUNNING_TIMEOUT = 9;
const int CBS_ERROR_USB_DOWN = 10;

const int STATE_IDLE = 0;
const int STATE_WORKING = 1;

const uint8_t BUF_FF[BLOCK_SIZE] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
const uint8_t BUF_IDENTIFICATION[BLOCK_SIZE] = {0x48, 0x0C, 0xE2, 0x53, 0x16, 0xF3, 0x8A};

LSM6 imu;

int state;
uint8_t buf[BLOCK_SIZE];

uint16_t readImuBlockCount = 0U;
unsigned long lastReadFifoData = 0UL;
unsigned long lastKeepRunningCmd = 0UL;

///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
inline void calcCrc(uint8_t * buf)
{
  //FIXME should use CRC
  buf[6] =  buf[0] + buf[1] + buf[2] + buf[3] + buf[4] + buf[5];
}

byte CRC8(const byte *data, byte len) {
  byte crc = 0x00;
  while (len--) {
    byte extract = *data++;
    for (byte tempI = 8; tempI; tempI--) {
      byte sum = (crc ^ extract) & 0x01;
      crc >>= 1;
      if (sum) {
        crc ^= 0x8C;
      }
      extract >>= 1;
    }
  }
  return crc;
}

inline void writeCB (uint8_t status)
{
  buf[0] = status;
  buf[1] = 0;
  buf[2] = 0;
  buf[3] = 0;
  buf[4] = 0;
  buf[5] = 0;
  buf[6] = status;

  SerialUSB.write(BUF_FF, BLOCK_SIZE);
  SerialUSB.write(buf, BLOCK_SIZE);
}

void setup()
{
  delay(100);
  SerialUSB.begin(9600);
  Wire.begin();
  Wire.setClock(3400000);
  pinMode(LED_PIN, OUTPUT);
  imu.init();

  state = STATE_IDLE;
}

void imuReset()
{
  imu.writeReg(imu.CTRL1_XL, 0xA0);
  imu.writeReg(imu.CTRL3_C, 0x01);
  delay(50); //just in case - reset takes 50 μs
  digitalWrite(LED_PIN, LOW);
}

void imuPowerDown()
{
  imu.writeReg(imu.CTRL1_XL, 0x00);
  imu.writeReg(imu.FIFO_CTRL5, 0x00);
  delay(50); //just in case
  digitalWrite(LED_PIN, LOW);
}

//odr      XL_HM_MODE = 1               XL_HM_MODE = 0
//0 0 0 0  Power-down                   Power-down
//0 0 0 1  12.5 Hz (low power)          12.5 Hz (high performance)
//0 0 1 0  26 Hz (low power)            26 Hz (high performance)
//0 0 1 1  52 Hz (low power)            52 Hz (high performance)
//0 1 0 0  104 Hz (normal mode)         104 Hz (high performance)
//0 1 0 1  208 Hz (normal mode)         208 Hz (high performance)
//0 1 1 0  416 Hz (high performance)    416 Hz (high performance)
//0 1 1 1  833 Hz (high performance)    833 Hz (high performance)
//1 0 0 0  1.66 kHz (high performance)  1.66 kHz (high performance)
//1 0 0 1  3.33 kHz (high performance)  3.33 kHz (high performance)
//1 0 1 0  6.66 kHz (high performance)  6.66 kHz (high performance)
//
//fs
//00: ±2 g; 01: ±16 g; 10: ±4 g; 11: ±8 g)
void imuStart(uint8_t xl_odr, uint8_t xl_fs)
{
  digitalWrite(LED_PIN, HIGH);

  uint8_t xl_bw = 0x0;
  imu.writeReg(imu.CTRL1_XL,
               ((xl_odr & 0x0F)) << 4 |
               ((xl_fs & 0x03)) << 2 |
               ((xl_bw & 0x03)));

  delay(50); //skip some samples (see Application note - Table 17. Accelerometer number of samples to be discarded)

  uint16_t fifo_thr = 100; //FIFO threshold level setting (for watermark flag)
  imu.writeReg(imu.FIFO_CTRL1, fifo_thr & 0x00FF);
  imu.writeReg(imu.FIFO_CTRL2, (fifo_thr & 0x0F00) >> 8);

  uint8_t fifo_dec_xl = 0x1; //Accelerometer FIFO decimation setting - No decimation
  imu.writeReg(imu.FIFO_CTRL3, fifo_dec_xl & 0x07);

  uint8_t fifo_odr = xl_odr;
  uint8_t fifo_mode = 0x6; //FIFO mode selection - Continuous mode. If the FIFO is full, the new sample overwrites the older one.
  imu.writeReg(imu.FIFO_CTRL5,
               ((fifo_odr & 0x0F)) << 3 |
               ((fifo_mode & 0x07)) );
}

void printFifoStatus()
{
  uint8_t s1 = imu.readReg(imu.FIFO_STATUS1);
  uint8_t s2 = imu.readReg(imu.FIFO_STATUS2);
  uint8_t s3 = imu.readReg(imu.FIFO_STATUS3);
  uint8_t s4 = imu.readReg(imu.FIFO_STATUS4);

  if (s2 & 0x80) //FIFO watermark status
    SerialUSB.print("T ");
  else
    SerialUSB.print("t ");

  if (s2 & 0x40) //FIFO overrun status
    SerialUSB.print("O ");
  else
    SerialUSB.print("o ");

  if (s2 & 0x20) //FIFO full status
    SerialUSB.print("F ");
  else
    SerialUSB.print("f ");

  if (s2 & 0x10) //FIFO empty bit
    SerialUSB.print("E ");
  else
    SerialUSB.print("e ");

  uint16_t fifosize = s1 | ((s2 & 0x0F) << 8);
  uint16_t pattern = s3 | ((s4 & 0x03) << 8);
  SerialUSB.print(fifosize);
  SerialUSB.print(" ");
  SerialUSB.print(pattern);

  if (pattern != 0)
    SerialUSB.println("Pattern failure!!!");

  if (pattern != 0 || s2 & 0x60)
    delay(20000);
}

boolean isFifoPatternOk() {
  uint8_t s3 = imu.readReg(imu.FIFO_STATUS3);
  uint8_t s4 = imu.readReg(imu.FIFO_STATUS4);

  uint16_t pattern = s3 | ((s4 & 0x03) << 8);

  return pattern == 0;
}

// return:
//   <0 error
//    0 pull no data
//   >0 count of read measurements
int readXl()
{
  uint8_t s2 = imu.readReg(imu.FIFO_STATUS2); //FIXME no communication error check
  if ((s2 & 0x60) != 0)
    return -CBS_ERROR_IMU_FIFO_OVERFLOW;

  if ((s2 & 0x80) == 0)
    return 0;

  Wire.beginTransmission(imu.address);
  Wire.write(imu.FIFO_DATA_OUT_L);
  Wire.endTransmission(false);

  Wire.requestFrom(imu.address, (uint8_t)(6 * BLOCKS_TO_READ));
  if (Wire.available() != (6 * BLOCKS_TO_READ))
    return -CBS_ERROR_IMU_FIFO_READ_LESS_THAN_REQUESTED;

  for (int i = 0; i < BLOCKS_TO_READ; i++)
  {
    buf[0] = Wire.read(); //xla
    buf[1] = Wire.read(); //xha
    buf[2] = Wire.read(); //yla
    buf[3] = Wire.read(); //yha
    buf[4] = Wire.read(); //zla
    buf[5] = Wire.read(); //zha

    // combine high and low bytes
    //int16_t x = (int16_t)(buf[1] << 8 | buf[0]);
    //int16_t y = (int16_t)(buf[3] << 8 | buf[2]);
    //int16_t z = (int16_t)(buf[5] << 8 | buf[4]);
    //snprintf(report, sizeof(report), "%d %d %d \n",
    //        x, y, z);

    calcCrc(buf);
    SerialUSB.write(buf, BLOCK_SIZE);
  }

  return BLOCKS_TO_READ;
}

// return:
//   <0 error
//    0 ok
int readChunkXl()
{
  int result;
  uint32_t pullsWoData = 0;
  unsigned long now, started;

  now = started = millis();
  while ((now - started) < READ_CHUNK_PERIOD) {
    result = readXl();
    now = millis();
    if (result < 0) {
      return result;
    }
    if (result == 0) {
      if ((now - lastReadFifoData) > NO_DATA_TIMEOUT) {
        return -CBS_ERROR_IMU_FIFO_NO_DATA_TIMEOUT;
      }
      pullsWoData++;
    }
    else {
      lastReadFifoData = now;
      readImuBlockCount += result;
    }

    if(!SerialUSB.dtr())
      return -CBS_ERROR_USB_DOWN;
  }

  if (!isFifoPatternOk()) {
    return -CBS_ERROR_IMU_FIFO_PATTERN;
  }

  buf[0] = CBS_STATISTICS;
  buf[1] = (uint8_t) (pullsWoData);
  buf[2] = (uint8_t) (pullsWoData >> 8);
  buf[3] = (uint8_t) (pullsWoData >> 16);
  buf[4] = (uint8_t) (readImuBlockCount);
  buf[5] = (uint8_t) (readImuBlockCount >> 8);
  calcCrc(buf);
  SerialUSB.write(BUF_FF, BLOCK_SIZE);
  SerialUSB.write(buf, BLOCK_SIZE);

  return 0;
}

void loop() {

  // Command byte
  //
  // MSB     LSB
  // CC FF FF SS
  //
  // CC   command COMMAND_*
  // FFFF odr (only for COMMAND_START_READING_IMU)
  // SS   full scale (only for COMMAND_START_READING_IMU)

  int cmd_byte = SerialUSB.read();
  if (cmd_byte != -1)
  {
    int cmd_code = cmd_byte >> 6;
    uint8_t xl_odr = (cmd_byte >> 2) & 0x0F;
    uint8_t xl_fs = cmd_byte & 0x03;

    switch (cmd_code) {

      case COMMAND_STOP_READING_IMU:
        if (state == STATE_WORKING) {
          imuPowerDown();
          state = STATE_IDLE;
          writeCB(CBS_END_DATA_STREAM);
        }
        else
          writeCB(CBS_ERROR_WRONG_COMMAND);
        break;

      case COMMAND_START_READING_IMU:
        if (state == STATE_WORKING) {
          imuPowerDown();
          state = STATE_IDLE;
          writeCB(CBS_ERROR_WRONG_COMMAND);
        }
        else if (xl_odr >= 1 && xl_odr <= 10)
        {
          SerialUSB.write(BUF_IDENTIFICATION, BLOCK_SIZE);
          uint8_t whoAmI = imu.readReg(imu.WHO_AM_I);
          if (whoAmI != 0x69)
            writeCB(CBS_ERROR_IMU_NOT_FOUND);
          else {
            imuReset();
            imuStart(xl_odr, xl_fs);
            lastReadFifoData = lastKeepRunningCmd = millis();
            readImuBlockCount = 0;
            state = STATE_WORKING;
            writeCB(CBS_START_DATA_STREAM);
          }
        }
        else
          writeCB(CBS_ERROR_WRONG_COMMAND);
        break;

      case COMMAND_KEEP_READING_IMU:
        if (state == STATE_WORKING)
          lastKeepRunningCmd = millis();
        else
          writeCB(CBS_ERROR_WRONG_COMMAND);
        break;

      case COMMAND_GET_IDENTIFICATION_CODE:
        if (state == STATE_WORKING) {
          imuPowerDown();
          state = STATE_IDLE;
          writeCB(CBS_ERROR_WRONG_COMMAND);
        }
        else
          SerialUSB.write(BUF_IDENTIFICATION, BLOCK_SIZE);
        break;
    }
  }

  switch (state) {
    case STATE_IDLE:
      delay(50);
      break;
    case STATE_WORKING:
      int result = readChunkXl();
      if (result < 0) {
        imuReset();
        state = STATE_IDLE;
        writeCB(-result);
      }
      else if ((millis() - lastKeepRunningCmd) > KEEP_RUNNING_TIMEOUT ) {
        imuPowerDown();
        state = STATE_IDLE;
        writeCB(CBS_ERROR_KEEP_RUNNING_TIMEOUT);
      }
      break;
  }
}
